# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午6:51
@file: transform.py
@author: zj
@description: Transform settings
"""

from yacs.config import CfgNode as CN


def add_config(_C: CN) -> None:
    # ---------------------------------------------------------------------------- #
    # Transform
    # ---------------------------------------------------------------------------- #
    _C.TRANSFORM = CN()
    # _C.TRANSFORM.TRAIN_METHODS = ('Resize', 'CenterCrop', 'ToTensor', 'Normalize')
    # _C.TRANSFORM.TEST_METHODS = ('Resize', 'CenterCrop', 'ToTensor', 'Normalize')
    _C.TRANSFORM.TRAIN_METHODS = ('RandomResizedCrop', 'RandomHorizontalFlip')
    _C.TRANSFORM.TEST_METHODS = ('Resize', 'CenterCrop')

    # ConvertImageDtype(dtype)
    # dtype: ['uint8', 'float32']
    _C.TRANSFORM.ConvertImageDtype = 'uint8'

    # Normalize(mean, std, inplace=False)
    # Default using Mean and STD calculated using Imagenet
    # Args:
    #     mean (sequence): Sequence of means for each channel.
    #     std (sequence): Sequence of standard deviations for each channel.
    #     inplace(bool,optional): Bool to make this operation in-place.
    _C.TRANSFORM.NORMALIZE = ((0.485, 0.456, 0.406), (0.229, 0.224, 0.225), False)

    # ---------------------------------------------------------------------------- #
    # Augment transform
    # ---------------------------------------------------------------------------- #
    # AutoAugment(policy, interpolation)
    # policy: ['IMAGENET', 'CIFAR10', 'SVHN']
    # interpolation: ['NEAREST', 'BILINEAR', 'BICUBIC']
    _C.TRANSFORM.AutoAugment = ("IMAGENET", "NEAREST")

    # ---------------------------------------------------------------------------- #
    # Geometric transform
    # ---------------------------------------------------------------------------- #
    # Desired output size of the crop.
    # If size is an int instead of sequence like (h, w), a square crop (size, size) is made.
    # If provided a sequence of length 1, it will be interpreted as (size[0], size[0]).
    _C.TRANSFORM.TRAIN_CROP = (224,)
    _C.TRANSFORM.TEST_CROP = (224,)

    # RandomHorizontalFlip(p=0.5)
    # Args:
    #     p (float): probability of the image being flipped. Default value is 0.5
    _C.TRANSFORM.RandomHorizontalFlip = 0.5

    # RandomVerticalFlip(p=0.5)
    # Args:
    #     p (float): probability of the image being flipped. Default value is 0.5
    _C.TRANSFORM.RandomVerticalFlip = 0.5

    # RandomRotate(degrees, interpolation=InterpolationMode.NEAREST, expand=False)
    # interpolation: ['NEAREST', 'BILINEAR', 'BICUBIC']
    # Args:
    #    degrees (sequence or number): Range of degrees to select from.
    #         If degrees is a number instead of sequence like (min, max), the range of degrees
    #         will be (-degrees, +degrees).
    #     interpolation (InterpolationMode): Desired interpolation enum defined by
    #         :class:`torchvision.transforms.InterpolationMode`. Default is ``InterpolationMode.NEAREST``.
    #         If input is Tensor, only ``InterpolationMode.NEAREST``, ``InterpolationMode.BILINEAR`` are supported.
    #         For backward compatibility integer values (e.g. ``PIL.Image.NEAREST``) are still acceptable.
    #     expand (bool, optional): Optional expansion flag.
    #         If true, expands the output to make it large enough to hold the entire rotated image.
    #         If false or omitted, make the output image the same size as the input image.
    #         Note that the expand flag assumes rotation around the center and no translation.
    _C.TRANSFORM.RandomRotate = (0, "NEAREST", False)

    # If size is a sequence like (h, w), output size will be matched to this.
    # If size is an int, smaller edge of the image will be matched to this number.
    # i.e, if height > width, then image will be rescaled to (size * height / width, size).
    _C.TRANSFORM.TRAIN_RESIZE = (224,)
    _C.TRANSFORM.TEST_RESIZE = (256,)

    # Desired output size of the crop.
    # If size is an int instead of sequence like (h, w), a square crop (size, size) is made.
    # If provided a sequence of length 1, it will be interpreted as (size[0], size[0]).
    _C.TRANSFORM.TRAIN_RESIZE_CROP = (224,)
    _C.TRANSFORM.TEST_RESIZE_CROP = (224,)

    # ---------------------------------------------------------------------------- #
    # Color transform
    # ---------------------------------------------------------------------------- #
    # ColorJitter(brightness=0, contrast=0, saturation=0, hue=0)
    # brightness (float or tuple of float (min, max)): How much to jitter brightness.
    #     brightness_factor is chosen uniformly from [max(0, 1 - brightness), 1 + brightness]
    #     or the given [min, max]. Should be non negative numbers.
    # contrast (float or tuple of float (min, max)): How much to jitter contrast.
    #     contrast_factor is chosen uniformly from [max(0, 1 - contrast), 1 + contrast]
    #     or the given [min, max]. Should be non negative numbers.
    # saturation (float or tuple of float (min, max)): How much to jitter saturation.
    #     saturation_factor is chosen uniformly from [max(0, 1 - saturation), 1 + saturation]
    #     or the given [min, max]. Should be non negative numbers.
    # hue (float or tuple of float (min, max)): How much to jitter hue.
    #     hue_factor is chosen uniformly from [-hue, hue] or the given [min, max].
    #     Should have 0<= hue <= 0.5 or -0.5 <= min <= max <= 0.5.
    _C.TRANSFORM.ColorJitter = (0, 0, 0, 0)

    # RandomAutocontrast(p=0.5)
    # p (float): probability of the image being autocontrasted. Default value is 0.5
    _C.TRANSFORM.RandomAutocontrast = 0.5

    # RandomAdjustSharpness(sharpness_factor, p=0.5)
    # sharpness_factor (float):  How much to adjust the sharpness. Can be
    #    any non negative number. 0 gives a blurred image, 1 gives the
    #    original image while 2 increases the sharpness by a factor of 2.
    # p (float): probability of the image being color inverted. Default value is 0.5
    _C.TRANSFORM.RandomAdjustSharpness = (1, 0.5)

    # RandomErasing(p=0.5, scale=(0.02, 0.33), ratio=(0.3, 3.3), value=0, inplace=False)
    # Args:
    #      p: probability that the random erasing operation will be performed.
    #      scale: range of proportion of erased area against input image.
    #      ratio: range of aspect ratio of erased area.
    #      value: erasing value. Default is 0. If a single int, it is used to
    #         erase all pixels. If a tuple of length 3, it is used to erase
    #         R, G, B channels respectively.
    #         If a str of 'random', erasing each pixel with random values.
    #      inplace: boolean to make this transform inplace. Default set to False.
    _C.TRANSFORM.RandomErasing = (0.5, (0.02, 0.33), (0.3, 3.3), 0, False)

    # RandomPosterize(bits, p=0.5)
    # Args:
    #     bits (int): number of bits to keep for each channel (0-8)
    #     p (float): probability of the image being color inverted. Default value is 0.5
    _C.TRANSFORM.RandomPosterize = (8, 0.5)



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午10:55
@file: util.py
@author: zj
@description: 
"""

import math

import torch.nn as nn
from typing import Optional


def create_linear(in_features: int, out_features: int, bias: Optional[bool] = True) -> nn.Module:
    fc = nn.Linear(in_features, out_features, bias=bias)
    reset_linear_parameters(fc)

    return fc


def reset_linear_parameters(module: nn.Module) -> None:
    """
    refer to: [[PyTorch]torch.nn各个层使用的默认的初始化分布](https://zhuanlan.zhihu.com/p/190207193)
    """
    assert isinstance(module, nn.Linear)
    nn.init.kaiming_uniform_(module.weight, a=math.sqrt(5))
    if module.bias is not None:
        fan_in, _ = nn.init._calculate_fan_in_and_fan_out(module.weight)
        bound = 1 / math.sqrt(fan_in)
        nn.init.uniform_(module.bias, -bound, bound)



# -*- coding: utf-8 -*-

"""
@date: 2022/4/11 下午3:51
@file: square_pad.py
@author: zj
@description: 
"""
from typing import Optional

import numpy as np
from PIL.Image import Image

import torchvision.transforms.functional as F


class SquarePad(object):
    """
    By filling the shorter edges, the size of the image becomes square
    refer to [How to resize and pad in a torchvision.transforms.Compose()?](https://discuss.pytorch.org/t/how-to-resize-and-pad-in-a-torchvision-transforms-compose/71850)
    """

    def __call__(self, image: Image) -> Image:
        w, h = image.size
        max_wh = np.max([w, h])
        hp = int((max_wh - w) / 2)
        vp = int((max_wh - h) / 2)
        padding = [hp, vp, hp, vp]
        return F.pad(image, padding, 0, 'constant')



# -*- coding: utf-8 -*-

"""
@date: 2022/4/16 下午7:55
@file: mix_dataset.py
@author: zj
@description: CCCF is a custom mixed classification dataset
"""
from typing import Optional, Tuple, Any, Callable

import os
import numpy as np
from PIL import Image

from torch.utils.data import Dataset
from zcls2.config.key_word import KEY_SEP

__all__ = ['CCCF']


class CCCF(Dataset):
    """
    CCCF is a custom mixed classification dataset, including
    1. CIFAR100: https://paperswithcode.com/dataset/cifar-100
    2. CUB-200-2011: https://paperswithcode.com/dataset/cub-200-2011
    3. Caltech-101: https://paperswithcode.com/dataset/caltech-101
    4. Food-101: https://paperswithcode.com/dataset/food-101

    The classes num = 100 + 200 + 101 + 101 = 502
    """

    def __init__(self, root: str,
                 train: Optional[bool] = True,
                 transform: Optional[Callable] = None,
                 target_transform: Optional[Callable] = None) -> None:
        assert os.path.isdir(root), root

        class_path = os.path.join(root, 'classes.txt')
        assert os.path.isfile(class_path), class_path
        train_path = os.path.join(root, 'train.txt')
        assert os.path.isfile(train_path), train_path
        test_path = os.path.join(root, 'test.txt')
        assert os.path.isfile(test_path), test_path

        classes = np.loadtxt(class_path, dtype=str, delimiter=' ')
        data_list = np.loadtxt(train_path, dtype=str, delimiter=KEY_SEP) if train else \
            np.loadtxt(test_path, dtype=str, delimiter=KEY_SEP)

        self.classes = classes
        self.data = [img_path for img_path, target in data_list]
        self.targets = [int(target) for img_path, target in data_list]

        self.root = root
        self.transform = transform
        self.target_transform = target_transform

    def __getitem__(self, index) -> Tuple[Any, Any]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)

    def __repr__(self) -> str:
        return self.__class__.__name__ + ' (' + self.root + ')'



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午5:20
@file: ghostnet.py
@author: zj
@description: 
"""
from typing import Optional, Dict

from torch import Tensor, nn
import torch.nn.functional as F

import timm.models as models
from timm.models import GhostNet

from zcls2.config.key_word import KEY_OUTPUT
from .util import create_linear

__supported_model__ = ['ghostnet_050', 'ghostnet_100', 'ghostnet_130']


def forward(self: GhostNet, x: Tensor) -> Dict:
    x = self.forward_features(x)
    x = self.flatten(x)
    if self.dropout > 0.:
        x = F.dropout(x, p=self.dropout, training=self.training)
    x = self.classifier(x)
    return {
        KEY_OUTPUT: x
    }


def get_ghostnet(pretrained: Optional[bool] = False,
                 num_classes: Optional[int] = 1000,
                 arch: Optional[str] = 'ghostnet_050') -> nn.Module:
    assert arch in __supported_model__, f"{arch} not in {__supported_model__}"

    if pretrained == True:
        model = models.__dict__[arch](pretrained=True, num_classes=1000)
        if num_classes != 1000:
            old_fc = model.classifier
            assert isinstance(old_fc, nn.Linear)

            in_features = old_fc.in_features
            new_fc = create_linear(in_features, num_classes, bias=old_fc.bias is not None)

            model.classifier = None
            model.classifier = new_fc
    else:
        model = models.__dict__[arch](pretrained=False, num_classes=num_classes)

    model.forward = forward
    return model


if __name__ == '__main__':
    model = get_ghostnet(num_classes=501, arch='ghostnet_130', pretrained=True)
    print(model)



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午4:47
@file: build.py
@author: zj
@description: 
"""
from typing import Tuple, Any

from yacs.config import CfgNode

from torch.utils.data.distributed import DistributedSampler


def build_sampler(cfg: CfgNode, train_dataset, val_dataset) -> Tuple[Any, Any]:
    train_sampler = None
    val_sampler = None
    if cfg.DISTRIBUTED:
        train_sampler = DistributedSampler(train_dataset)
        val_sampler = DistributedSampler(val_dataset)

    return train_sampler, val_sampler



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:30
@file: build.py
@author: zj
@description: 
"""

from typing import Tuple

from yacs.config import CfgNode
from torch.utils.data import IterableDataset, DataLoader, Sampler

from .transform.build import build_transform
from .dataset.build import build_dataset
from .sampler.build import build_sampler
from .dataloader.build import build_dataloader


def build_data(cfg: CfgNode) -> Tuple[Sampler, DataLoader, DataLoader]:
    train_transform, train_target_transform = build_transform(cfg, is_train=True)
    train_dataset = build_dataset(cfg, train_transform, train_target_transform, is_train=True)

    val_transform, val_target_transform = build_transform(cfg, is_train=False)
    val_dataset = build_dataset(cfg, val_transform, val_target_transform, is_train=False)

    if isinstance(train_dataset, IterableDataset):
        train_sampler, val_sampler = None, None
        shuffle = False
    else:
        train_sampler, val_sampler = build_sampler(cfg, train_dataset, val_dataset)
        shuffle = train_sampler is None

    train_loader, val_loader = build_dataloader(cfg,
                                                train_dataset, val_dataset, train_sampler, val_sampler,
                                                shuffle)

    return train_sampler, train_loader, val_loader



# -*- coding: utf-8 -*-

"""
@date: 2022/4/4 下午10:24
@file: cosine_annealing_lr.py
@author: zj
@description: 
"""

from typing import Optional

from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import CosineAnnealingLR


def build_cosine_annearling_lr(optimizer: Optimizer,
                               warmup: Optional[bool] = True,
                               warmup_epoch: Optional[int] = 5,
                               max_epoch: Optional[int] = 90,
                               minimal_lr: Optional[float] = 1e-6) -> CosineAnnealingLR:
    assert isinstance(optimizer, Optimizer)

    if warmup:
        max_epoch = max_epoch - warmup_epoch
    # Ensure that the last round is minimal_lr
    lr_scheduler = CosineAnnealingLR(optimizer, T_max=max_epoch - 1, eta_min=minimal_lr)

    return lr_scheduler



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:41
@file: misc.py
@author: zj
@description: 
"""

import torch
import random

import numpy as np


def to_python_float(t):
    if hasattr(t, 'item'):
        return t.item()
    else:
        return t[0]


def init_seed(seed=0):
    """
    Same as Apex settings
    See
    1. [REPRODUCIBILITY](https://pytorch.org/docs/stable/notes/randomness.html)
    2. [PyTorch设置随机种子](https://blog.csdn.net/weixin_41978699/article/details/121312297)
    """
    # random.seed(seed)
    # np.random.seed(seed)
    torch.manual_seed(seed)
    # torch.cuda.manual_seed(seed)
    # torch.cuda.manual_seed_all(seed)



# -*- coding: utf-8 -*-

"""
@date: 2022/4/4 下午10:42
@file: multi_step_lr.py
@author: zj
@description: 
"""

from typing import Optional

from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import MultiStepLR


def build_multistep_lr(optimizer: Optimizer,
                       milestones: Optional[list] = None,
                       gamma: Optional[float] = 0.1) -> MultiStepLR:
    assert isinstance(optimizer, Optimizer)
    if milestones is None:
        milestones = [30, 60, 80]

    return MultiStepLR(optimizer, milestones=milestones, gamma=gamma)



# -*- coding: utf-8 -*-

"""
@date: 2022/4/10 上午10:51
@file: mobilenet.py
@author: zj
@description: Custom MobileNetV2/V3, derived from torchvision
"""
from typing import Any, List, Optional, Callable, Dict

from torch import nn, Tensor
from torchvision.models.mobilenet import MobileNetV2, MobileNetV3
from torchvision.models import mobilenetv2
from torchvision.models import mobilenetv3

# See vision/torchvision/_internally_replaced_utils.py
# https://github.com/pytorch/vision/blob/b50ffef5f85029b1440ac155ca1e6d95c55520aa/torchvision/_internally_replaced_utils.py
from torchvision.models.mobilenetv3 import InvertedResidualConfig

try:
    from torch.hub import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url

from zcls2.config.key_word import KEY_OUTPUT

__all__ = [
    'mobilenet_v2', 'mobilenet_v3_large', 'mobilenet_v3_small'
]


class ZMobileNetV2(MobileNetV2):

    def __init__(self, num_classes: int = 1000, width_mult: float = 1.0,
                 inverted_residual_setting: Optional[List[List[int]]] = None, round_nearest: int = 8,
                 block: Optional[Callable[..., nn.Module]] = None,
                 norm_layer: Optional[Callable[..., nn.Module]] = None) -> None:
        super().__init__(num_classes, width_mult, inverted_residual_setting, round_nearest, block, norm_layer)

    def forward(self, x: Tensor) -> Dict:
        res = super().forward(x)
        return {
            KEY_OUTPUT: res
        }


def mobilenet_v2(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> MobileNetV2:
    """
    Constructs a MobileNetV2 architecture from
    `"MobileNetV2: Inverted Residuals and Linear Bottlenecks" <https://arxiv.org/abs/1801.04381>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    model = ZMobileNetV2(**kwargs)
    if pretrained:
        state_dict = load_state_dict_from_url(mobilenetv2.model_urls['mobilenet_v2'],
                                              progress=progress)

        # If the number of model outputs is different from the model setting,
        # the corresponding pretraining model weight will not be loaded
        assert isinstance(model.classifier[1], nn.Linear)
        if model.classifier[1].out_features != 1000:
            state_dict.pop('classifier.1.weight')
            state_dict.pop('classifier.1.bias')

        ret = model.load_state_dict(state_dict, strict=False)
        assert set(ret.missing_keys) == {'classifier.1.weight', 'classifier.1.bias'}, \
            f'Missing keys when loading pretrained weights: {ret.missing_keys}'
    return model


class ZMobileNetV3(MobileNetV3):

    def __init__(self, inverted_residual_setting: List[InvertedResidualConfig], last_channel: int,
                 num_classes: int = 1000, block: Optional[Callable[..., nn.Module]] = None,
                 norm_layer: Optional[Callable[..., nn.Module]] = None, **kwargs: Any) -> None:
        super().__init__(inverted_residual_setting, last_channel, num_classes, block, norm_layer, **kwargs)

    def forward(self, x: Tensor) -> Dict:
        res = super().forward(x)
        return {
            KEY_OUTPUT: res
        }


def _mobilenet_v3_model(
        arch: str,
        inverted_residual_setting: List[mobilenetv3.InvertedResidualConfig],
        last_channel: int,
        pretrained: bool,
        progress: bool,
        **kwargs: Any
):
    model = ZMobileNetV3(inverted_residual_setting, last_channel, **kwargs)
    if pretrained:
        if mobilenetv3.model_urls.get(arch, None) is None:
            raise ValueError("No checkpoint is available for model type {}".format(arch))
        state_dict = load_state_dict_from_url(mobilenetv3.model_urls[arch], progress=progress)

        # If the number of model outputs is different from the model setting,
        # the corresponding pretraining model weight will not be loaded
        assert isinstance(model.classifier[3], nn.Linear)
        if model.classifier[3].out_features != 1000:
            state_dict.pop('classifier.3.weight')
            state_dict.pop('classifier.3.bias')

        ret = model.load_state_dict(state_dict, strict=False)
        assert set(ret.missing_keys) == {'classifier.3.weight', 'classifier.3.bias'}, \
            f'Missing keys when loading pretrained weights: {ret.missing_keys}'
    return model


def mobilenet_v3_large(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> MobileNetV3:
    """
    Constructs a large MobileNetV3 architecture from
    `"Searching for MobileNetV3" <https://arxiv.org/abs/1905.02244>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    arch = "mobilenet_v3_large"
    inverted_residual_setting, last_channel = mobilenetv3._mobilenet_v3_conf(arch, **kwargs)
    return _mobilenet_v3_model(arch, inverted_residual_setting, last_channel, pretrained, progress, **kwargs)


def mobilenet_v3_small(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> MobileNetV3:
    """
    Constructs a small MobileNetV3 architecture from
    `"Searching for MobileNetV3" <https://arxiv.org/abs/1905.02244>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    arch = "mobilenet_v3_small"
    inverted_residual_setting, last_channel = mobilenetv3._mobilenet_v3_conf(arch, **kwargs)
    return _mobilenet_v3_model(arch, inverted_residual_setting, last_channel, pretrained, progress, **kwargs)



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:34
@file: trainer.py
@author: zj
@description: 
"""

import time

from typing import Optional
from yacs.config import CfgNode

import torch
import torch.nn as nn
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader

try:
    from apex.parallel import DistributedDataParallel as DDP
    from apex.fp16_utils import *
    from apex import amp, optimizers
    from apex.multi_tensor_apply import multi_tensor_applier
except ImportError:
    raise ImportError("Please install apex from https://www.github.com/nvidia/apex to run this example.")

from ..config.key_word import KEY_OUTPUT
from ..util.meter import AverageMeter
from ..util.prefetcher import data_prefetcher
from ..util.metric import accuracy
from ..util.distributed import reduce_tensor
from ..util.misc import to_python_float
from ..optim.lr_scheduler.build import adjust_learning_rate

from zcls2.util import logging

logger = logging.get_logger(__name__)


def train(cfg: CfgNode, train_loader: DataLoader, model: nn.Module, criterion: nn.Module, optimizer: Optimizer,
          epoch: Optional[int] = 1) -> None:
    batch_time = AverageMeter()
    losses = AverageMeter()
    top_k = cfg.TRAIN.TOP_K
    top_list = [AverageMeter() for _ in top_k]

    # switch to train mode
    model.train()
    end = time.time()

    warmup = cfg.LR_SCHEDULER.IS_WARMUP
    warmup_epoch = cfg.LR_SCHEDULER.WARMUP_EPOCH

    prefetcher = data_prefetcher(cfg, train_loader)
    input, target = prefetcher.next()
    i = 0
    while input is not None:
        i += 1
        if cfg.PROF >= 0 and i == cfg.PROF:
            logger.info("Profiling begun at iteration {}".format(i))
            torch.cuda.cudart().cudaProfilerStart()

        if cfg.PROF >= 0: torch.cuda.nvtx.range_push("Body of iteration {}".format(i))

        if warmup and epoch < warmup_epoch + 1:
            adjust_learning_rate(cfg, optimizer, epoch, i, len(train_loader))

        # compute output
        if cfg.PROF >= 0: torch.cuda.nvtx.range_push("forward")
        output = model(input)
        if cfg.PROF >= 0: torch.cuda.nvtx.range_pop()
        loss = criterion(output, target)

        # compute gradient and do SGD step
        optimizer.zero_grad()

        if cfg.PROF >= 0: torch.cuda.nvtx.range_push("backward")
        with amp.scale_loss(loss, optimizer) as scaled_loss:
            scaled_loss.backward()
        if cfg.PROF >= 0: torch.cuda.nvtx.range_pop()

        # for param in model.parameters():
        #     print(param.data.double().sum().item(), param.grad.data.double().sum().item())

        if cfg.PROF >= 0: torch.cuda.nvtx.range_push("optimizer.step()")
        optimizer.step()
        if cfg.PROF >= 0: torch.cuda.nvtx.range_pop()

        if i % cfg.PRINT_FREQ == 0:
            # Every print_freq iterations, check the loss, accuracy, and speed.
            # For best performance, it doesn't make sense to print these metrics every
            # iteration, since they incur an allreduce and some host<->device syncs.

            # Measure accuracy
            prec_list = accuracy(output[KEY_OUTPUT].data, target, topk=top_k)

            # Average loss and accuracy across processes for logging
            if cfg.DISTRIBUTED:
                reduced_loss = reduce_tensor(cfg.NUM_GPUS, loss.data)
                prec_list = [reduce_tensor(cfg.NUM_GPUS, prec) for prec in prec_list]
            else:
                reduced_loss = loss.data

            # to_python_float incurs a host<->device sync
            losses.update(to_python_float(reduced_loss), input.size(0))
            for i, prec in enumerate(prec_list):
                top_list[i].update(to_python_float(prec), input.size(0))

            torch.cuda.synchronize()
            batch_time.update((time.time() - end) / cfg.PRINT_FREQ)
            end = time.time()

            if cfg.RANK_ID == 0:
                logger_str = 'Epoch: [{0}/{1}][{2}/{3}] ' \
                             'Time {batch_time.val:.3f} ({batch_time.avg:.3f}) ' \
                             'Speed {4:.3f} ({5:.3f}) ' \
                             'Lr {lr:.10f} ' \
                             'Loss {loss.val:.10f} ({loss.avg:.4f}) '.format(
                    epoch, cfg.TRAIN.MAX_EPOCH, i, len(train_loader),
                    cfg.NUM_GPUS * cfg.DATALOADER.TRAIN_BATCH_SIZE / batch_time.val,
                    cfg.NUM_GPUS * cfg.DATALOADER.TRAIN_BATCH_SIZE / batch_time.avg,
                    batch_time=batch_time,
                    lr=optimizer.param_groups[0]['lr'],
                    loss=losses)
                for k, top in zip(top_k, top_list):
                    logger_str += f'Prec@{k} {top.val:.3f} ({top.avg:.3f}) '
                logger.info(logger_str)
        if cfg.PROF >= 0: torch.cuda.nvtx.range_push("prefetcher.next()")
        input, target = prefetcher.next()
        if cfg.PROF >= 0: torch.cuda.nvtx.range_pop()

        # Pop range "Body of iteration {}".format(i)
        if cfg.PROF >= 0: torch.cuda.nvtx.range_pop()

        if cfg.PROF >= 0 and i == cfg.PROF + 10:
            logger.info("Profiling ended at iteration {}".format(i))
            torch.cuda.cudart().cudaProfilerStop()
            quit()



# -*- coding: utf-8 -*-

"""
@date: 2022/4/4 下午10:13
@file: large_margin_softmax_loss.py
@author: zj
@description: 
"""

import torch
import torch.nn as nn

'''
proposed in the BMVC2019 paper: [Large Margin in Softmax Cross-Entropy Loss
link to paper](https://staff.aist.go.jp/takumi.kobayashi/publication/2019/BMVC2019.pdf)
'''


##
# version 1: use torch.autograd
class LargeMarginSoftmaxV1(nn.Module):

    def __init__(self, lam=0.3, reduction='mean', ignore_index=255):
        super(LargeMarginSoftmaxV1, self).__init__()
        self.reduction = reduction
        self.ignore_index = ignore_index
        self.lam = lam
        self.ce_crit = nn.CrossEntropyLoss(
            reduction='none', ignore_index=ignore_index)

    def __call__(self, logits, label):
        # def forward(self, logits, label):
        '''
        Same usage method as nn.CrossEntropyLoss:
            >>> criteria = LargeMarginSoftmaxV1()
            >>> logits = torch.randn(8, 19, 384, 384) # nchw, float/half
            >>> lbs = torch.randint(0, 19, (8, 384, 384)) # nhw, int64_t
            >>> loss = criteria(logits, lbs)
        '''
        # overcome ignored label
        logits = logits.float()
        if logits.requires_grad is True:
            logits.retain_grad()
            logits.register_hook(lambda grad: grad)
        with torch.no_grad():
            num_classes = logits.size(1)
            coeff = 1. / (num_classes - 1.)
            lb = label.clone().detach()
            mask = label == self.ignore_index
            lb[mask] = 0
            idx = torch.zeros_like(logits).scatter_(1, lb.unsqueeze(1), 1.)

        lgts = logits - idx * 1.e6
        q = lgts.softmax(dim=1)
        q = q * (1. - idx)

        log_q = lgts.log_softmax(dim=1)
        log_q = log_q * (1. - idx)
        mg_loss = ((q - coeff) * log_q) * (self.lam / 2)
        mg_loss = mg_loss * (1. - idx)
        mg_loss = mg_loss.sum(dim=1)

        ce_loss = self.ce_crit(logits, label)
        loss = ce_loss + mg_loss
        loss = loss[mask == 0]

        if self.reduction == 'mean':
            loss = loss.mean()
        if self.reduction == 'sum':
            loss = loss.sum()

        return loss


def build_large_margin_softmax_loss(reduction: str = 'mean') -> nn.Module:
    return LargeMarginSoftmaxV1(reduction=reduction)



# -*- coding: utf-8 -*-

"""
@date: 2020/11/10 下午5:02
@file: cifar.py
@author: zj
@description: 
"""

from typing import Optional, Tuple, Any, Callable

from torch.utils.data import Dataset
from torchvision.datasets import CIFAR10, CIFAR100

__all__ = ['CIFAR']


class CIFAR(Dataset):

    def __init__(self, root: str,
                 train: Optional[bool] = True,
                 transform: Optional[Callable] = None,
                 target_transform: Optional[Callable] = None,
                 is_cifar100: Optional[bool] = True) -> None:
        if is_cifar100:
            self.data_set = CIFAR100(root, train=train, transform=transform, target_transform=target_transform,
                                     download=True)
        else:
            self.data_set = CIFAR10(root, train=train, transform=transform, target_transform=target_transform,
                                    download=True)
        self.classes = self.data_set.classes
        self.root = root

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        return self.data_set.__getitem__(index)

    def __len__(self) -> int:
        return self.data_set.__len__()

    def __repr__(self) -> str:
        return self.__class__.__name__ + ' (' + self.root + ')'



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:39
@file: distributed.py
@author: zj
@description: 
"""

import os

import torch
import torch.distributed as dist
import torch.backends.cudnn as cudnn

from yacs.config import CfgNode
from argparse import Namespace

from .misc import init_seed


def init_dist(args: Namespace, cfg: CfgNode) -> None:
    cudnn.benchmark = False if args.deterministic else True
    cudnn.deterministic = True if args.deterministic else False
    if args.deterministic:
        init_seed(args.local_rank)
        torch.set_printoptions(precision=10)

    args.distributed = False
    if 'WORLD_SIZE' in os.environ:
        args.distributed = int(os.environ['WORLD_SIZE']) > 1

    args.gpu = 0
    args.world_size = 1

    if args.distributed:
        args.gpu = args.local_rank
        torch.cuda.set_device(args.gpu)
        torch.distributed.init_process_group(backend=cfg.DIST_BACKEND, init_method=cfg.INIT_METHOD)
        args.world_size = torch.distributed.get_world_size()

    assert torch.backends.cudnn.enabled, "Amp requires cudnn backend to be enabled."


def reduce_tensor(world_size, tensor):
    rt = tensor.clone()
    dist.all_reduce(rt, op=dist.reduce_op.SUM)
    rt /= world_size
    return rt


def is_master_proc():
    """
    Determines if the current process is the master process.
    """
    if torch.distributed.is_initialized():
        return dist.get_rank() % get_world_size() == 0
    else:
        return True


def get_world_size():
    """
    Get the size of the world.
    """
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1
    return dist.get_world_size()



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:37
@file: build.py
@author: zj
@description: 
"""

from yacs.config import CfgNode
from torch.optim.optimizer import Optimizer

from .multi_step_lr import build_multistep_lr
from .cosine_annealing_lr import build_cosine_annearling_lr

__supported_lr_scheduler__ = [
    'MultiStepLR',
    'CosineAnnealingLR'
]


def adjust_learning_rate(cfg: CfgNode, optimizer: Optimizer, epoch: int, step: int, len_epoch: int) -> None:
    """LR schedule that should yield 76% converged accuracy with batch size 256"""
    lr = cfg.OPTIMIZER.LR

    warmup_epoch = cfg.LR_SCHEDULER.WARMUP_EPOCH
    """Warmup"""
    if epoch < warmup_epoch + 1:
        lr = lr * float(1 + step + (epoch - 1) * len_epoch) / (warmup_epoch * len_epoch)

    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def build_lr_scheduler(cfg: CfgNode, optimizer: Optimizer):
    lr_scheduler_name = cfg.LR_SCHEDULER.NAME
    assert lr_scheduler_name in __supported_lr_scheduler__

    if lr_scheduler_name == 'MultiStepLR':
        milestones = cfg.LR_SCHEDULER.MULTISTEP_LR.MILESTONES
        gamma = cfg.LR_SCHEDULER.MULTISTEP_LR.GAMMA
        return build_multistep_lr(optimizer, milestones=milestones, gamma=gamma)
    elif lr_scheduler_name == 'CosineAnnealingLR':
        warmup = cfg.LR_SCHEDULER.IS_WARMUP
        warmup_epoch = cfg.LR_SCHEDULER.WARMUP_EPOCH
        max_epoch = cfg.TRAIN.MAX_EPOCH

        return build_cosine_annearling_lr(optimizer,
                                          warmup=warmup, warmup_epoch=warmup_epoch, max_epoch=max_epoch,
                                          minimal_lr=1e-6)
    else:
        raise ValueError(f"{lr_scheduler_name} does not support")



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:54
@file: build.py
@author: zj
@description: 
"""

import torch.nn as nn
from yacs.config import CfgNode

from .cross_entropy_loss import build_cross_entropy_loss
from .large_margin_softmax_loss import build_large_margin_softmax_loss

__supported_criterion__ = [
    'CrossEntropyLoss',
    'LargeMarginSoftmaxV1'
]


def build_criterion(cfg: CfgNode) -> nn.Module:
    loss_name = cfg.MODEL.CRITERION.NAME
    reduction = cfg.MODEL.CRITERION.REDUCTION

    assert loss_name in __supported_criterion__

    if loss_name == 'CrossEntropyLoss':
        return build_cross_entropy_loss(reduction=reduction)
    elif loss_name == 'LargeMarginSoftmaxV1':
        return build_large_margin_softmax_loss(reduction=reduction)
    else:
        raise ValueError(f"{loss_name} does not support")



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:34
@file: prefetcher.py
@author: zj
@description: 
"""

import torch


class data_prefetcher():
    def __init__(self, cfg, loader):
        self.loader = iter(loader)
        self.stream = torch.cuda.Stream()

        mean, std, _ = cfg.TRANSFORM.NORMALIZE
        assert len(mean) == len(std)

        mean_list = list()
        std_list = list()
        for m, s in zip(mean, std):
            mean_list.append(m * 255)
            std_list.append(s * 255)
        n = len(mean_list)
        self.mean = torch.tensor(mean_list).cuda().view(1, n, 1, 1)
        self.std = torch.tensor(std_list).cuda().view(1, n, 1, 1)

        # self.mean = torch.tensor([0.485 * 255, 0.456 * 255, 0.406 * 255]).cuda().view(1, 3, 1, 1)
        # self.std = torch.tensor([0.229 * 255, 0.224 * 255, 0.225 * 255]).cuda().view(1, 3, 1, 1)
        # With Amp, it isn't necessary to manually convert data to half.
        # if args.fp16:
        #     self.mean = self.mean.half()
        #     self.std = self.std.half()
        self.preload()

    def preload(self):
        try:
            self.next_input, self.next_target = next(self.loader)
        except StopIteration:
            self.next_input = None
            self.next_target = None
            return
        # if record_stream() doesn't work, another option is to make sure device inputs are created
        # on the main stream.
        # self.next_input_gpu = torch.empty_like(self.next_input, device='cuda')
        # self.next_target_gpu = torch.empty_like(self.next_target, device='cuda')
        # Need to make sure the memory allocated for next_* is not still in use by the main stream
        # at the time we start copying to next_*:
        # self.stream.wait_stream(torch.cuda.current_stream())
        with torch.cuda.stream(self.stream):
            self.next_input = self.next_input.cuda(non_blocking=True)
            self.next_target = self.next_target.cuda(non_blocking=True)
            # more code for the alternative if record_stream() doesn't work:
            # copy_ will record the use of the pinned source tensor in this side stream.
            # self.next_input_gpu.copy_(self.next_input, non_blocking=True)
            # self.next_target_gpu.copy_(self.next_target, non_blocking=True)
            # self.next_input = self.next_input_gpu
            # self.next_target = self.next_target_gpu

            # With Amp, it isn't necessary to manually convert data to half.
            # if args.fp16:
            #     self.next_input = self.next_input.half()
            # else:
            self.next_input = self.next_input.float()
            self.next_input = self.next_input.sub_(self.mean).div_(self.std)

    def next(self):
        torch.cuda.current_stream().wait_stream(self.stream)
        input = self.next_input
        target = self.next_target
        if input is not None:
            input.record_stream(torch.cuda.current_stream())
        if target is not None:
            target.record_stream(torch.cuda.current_stream())
        self.preload()
        return input, target



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:30
@file: parser.py
@author: zj
@description:
args仅保留apex特有的训练配置，其他的训练参数均包含在cfg文件中
cfg负责数据/模型/损失函数/优化器/学习率调度器以及其他组件配置
"""

import os

import argparse
from argparse import Namespace

from yacs.config import CfgNode


def parse() -> Namespace:
    parser = argparse.ArgumentParser(description='ZCls2 Training with Pytorch')
    parser.add_argument('-cfg',
                        "--config",
                        type=str,
                        default="",
                        metavar="CONFIG",
                        help="path to config file")

    parser.add_argument('--resume', default='', type=str, metavar='PATH',
                        help='path to latest checkpoint (default: none)')
    parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                        help='evaluate model on validation set')

    parser.add_argument('--deterministic', action='store_true')

    parser.add_argument("--local_rank", default=os.getenv('LOCAL_RANK', 0), type=int)

    parser.add_argument('--opt-level', type=str)
    parser.add_argument('--keep-batchnorm-fp32', type=str, default=None)
    parser.add_argument('--loss-scale', type=str, default=None)
    parser.add_argument('--channels-last', type=bool, default=False)
    args = parser.parse_args()

    return args


def load_cfg(args: Namespace, cfg: CfgNode):
    cfg.DISTRIBUTED = args.distributed
    cfg.RANK_ID = args.gpu
    cfg.NUM_GPUS = args.world_size

    cfg.DETERMINISTIC = args.deterministic
    cfg.RNG_SEED = args.local_rank

    cfg.RESUME = args.resume
    cfg.EVALUATE = args.evaluate

    cfg.CHANNELS_LAST = args.channels_last



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午1:40
@file: checkpoint.py
@author: zj
@description: 
"""

import os
import torch
import shutil

from zcls2.util import logging

logger = logging.get_logger(__name__)


def save_checkpoint(state, is_best, output_dir='outputs', filename='checkpoint.pth.tar'):
    save_path = os.path.join(output_dir, filename)
    logger.info(f"Save to {save_path}")
    torch.save(state, save_path)
    if is_best:
        best_path = os.path.join(output_dir, 'model_best.pth.tar')
        logger.info(f"Copy to {best_path}")
        shutil.copyfile(save_path, best_path)



# -*- coding: utf-8 -*-

"""
@date: 2022/4/3 下午5:00
@file: sgd.py
@author: zj
@description: 
"""
from typing import Optional

import torch
from torch.optim.optimizer import Optimizer


def build_sgd(groups: list,
              lr: Optional[float] = 1e-3,
              momentum: Optional[float] = 0.9,
              weight_decay: Optional[float] = 1e-4) -> Optimizer:
    return torch.optim.SGD(groups,
                           lr=lr,
                           momentum=momentum,
                           weight_decay=weight_decay)
